<?php

namespace Drupal\revision_graph\Controller;

use Drupal\Component\Utility\Xss;
use Drupal\Core\Database\Connection;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\node\Controller\NodeController;
use Drupal\node\NodeInterface;
use Drupal\node\NodeStorageInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Returns responses for Node routes.
 */
class RevisionGraphController implements ContainerInjectionInterface {

  use StringTranslationTrait;

  /**
   * {@inheritdoc}
   */
  public function __construct(
    private readonly DateFormatterInterface $date_formatter,
    private readonly RendererInterface $renderer,
    private readonly EntityRepositoryInterface $entityRepository,
    private readonly Connection $connection,
    private readonly EntityTypeManagerInterface $entityTypeManager,
    private readonly LanguageManagerInterface $languageManager) {}

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('date.formatter'),
      $container->get('renderer'),
      $container->get('entity.repository'),
      $container->get('database'),
      $container->get('entity_type.manager'),
      $container->get('language_manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function revisionOverview(NodeInterface $node) {
    $commits = [];
    $node = $this->entityRepository->getActive($node->getEntityTypeId(), $node->id());
    foreach (\Drupal::languageManager()->getLanguages() as $language) {
      $revisions = $this->revisionOverviewItems($node, $language->getId());
      foreach ($this->revisionOverviewAsFakeCommits($node, $language->getId(), $revisions) as $commit) {
        $commits[$commit['id']] = $commit;
      }
    }
    $previous_commit = NULL;
    ksort($commits);
    $commits = array_values($commits);
    $branch_specific_head_commit = [];
    foreach($commits as &$commit ) {
      if ($previous_commit) {
        if ($previous_commit['branch'] != $commit['branch'] && !empty($branch_specific_head_commit[$commit['branch']]) ) {
          $commit['parents'][] = $branch_specific_head_commit[$commit['branch']]['id'];
        }
        else {
          $commit['parents'][] = $previous_commit['id'];
        }
      }
      $branch_specific_head_commit[$commit['branch']] = $commit;
      $previous_commit = $commit;
    }
    $commits = array_reverse($commits);
    return [
      '#theme' => 'revision_graph',
      '#attached' => [
        'library' => 'revision_graph/revision_graph',
        'drupalSettings' => [
          'revision_graph' => [
            'commits' => $commits,
            'initialBranch' => !empty($commits[0]['branch']) ? $commits[0]['branch'] : $node->language()->getId(),
          ],
        ],
      ],
    ];
  }

  /**
   * Generates an overview table of older revisions of a node.
   *
   * @param \Drupal\node\NodeInterface $node
   *   A node object.
   *
   * @return array
   *   An array as expected by \Drupal\Core\Render\RendererInterface::render().
   */
  public function revisionOverviewItems(NodeInterface $node, $language_code = NULL) {
    // Always use the latest revision in the current content language to
    // determine if this node has translations. This supports showing the
    // correct translation revisions for translations that only have.
    // non-default revisions.
    $node = $this->entityRepository->getActive($node->getEntityTypeId(), $node->id());
    $langcode = $language_code ?: $node->language()->getId();
    $languages = $node->getTranslationLanguages();
    $has_translations = (count($languages) > 1);
    $node_storage = $this->entityTypeManager->getStorage('node');
    $rows = [];
    $current_revision_displayed = FALSE;

    foreach ($this->getRevisionIds($node, $node_storage) as $vid) {
      /** @var \Drupal\node\NodeInterface $revision */
      $revision = $node_storage->loadRevision($vid);
      // Only show revisions that are affected by the language that is being
      // displayed.
      if ($revision->hasTranslation($langcode) && $revision->getTranslation($langcode)->isRevisionTranslationAffected()) {
        // We treat also the latest translation-affecting revision as current
        // revision, if it was the default revision, as its values for the
        // current language will be the same of the current default revision in
        // this case.
        $is_current_revision = $revision->isDefaultRevision() || (!$current_revision_displayed && $revision->wasDefaultRevision());
        if ($is_current_revision) {
          $current_revision_displayed = TRUE;
        }

        $message = ['#markup' => $revision->revision_log->value, '#allowed_tags' => Xss::getHtmlTagList()];
        $row = [
          'is_current_revision' => $is_current_revision,
          'timestamp' => $revision->revision_timestamp->value,
          'author' => $revision->getRevisionUser()->label(),
          'message' => (string) $this->renderer->renderInIsolation($message),
        ];
        $links = [];
        $view_url = new Url('entity.node.revision', ['node' => $node->id(), 'node_revision' => $vid]);
        $links['view'] = [
          'label' => (string) $this->t('View'),
          'url' => $view_url->toString(),
        ];
        if (!$is_current_revision) {
          if ($revision->access('revert revision')) {

            if ($has_translations) {
              $revert_url = Url::fromRoute('node.revision_revert_translation_confirm', ['node' => $node->id(), 'node_revision' => $vid, 'langcode' => $langcode]);
            }
            else {
              $revert_url = Url::fromRoute('node.revision_revert_confirm', ['node' => $node->id(), 'node_revision' => $vid]);
            }
            $links['revert'] = [
              'label' => $vid < $node->getRevisionId() ? $this->t('Revert') : $this->t('Set as current revision'),
              'url' => $revert_url->toString(),
            ];
          }
          if ($revision->access('delete revision')) {
            $delete_url = Url::fromRoute('node.revision_delete_confirm', ['node' => $node->id(), 'node_revision' => $vid]);
            $links['delete'] = [
              'label' => $this->t('Delete'),
              'url' => $delete_url->toString(),
            ];
          }
        }
        $row['urls'] = array_values($links);
        $rows[$vid] = $row;
      }
    }
    return $rows;
  }


  /**
   * Gets a list of node revision IDs for a specific node.
   *
   * @param \Drupal\node\NodeInterface $node
   *   The node entity.
   * @param \Drupal\node\NodeStorageInterface $node_storage
   *   The node storage handler.
   *
   * @return int[]
   *   Node revision IDs (in descending order).
   */
  protected function getRevisionIds(NodeInterface $node, NodeStorageInterface $node_storage) {
    $result = $node_storage->getQuery()
      ->accessCheck(TRUE)
      ->allRevisions()
      ->condition($node->getEntityType()->getKey('id'), $node->id())
      ->sort($node->getEntityType()->getKey('revision'), 'DESC')
      ->execute();
    return array_keys($result);
  }

  /**
   * Load revision data as git commits.
   *
   * @param Drupal\node\NodeInterface $node
   *   Node for which we need to load the revision graph.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   *
   * @return array
   *   Revision data as list of commits.
   */
  protected function revisionOverviewAsFakeCommits(NodeInterface $node, $langcode, $revisions = []) {
    $tags_list = [
      'published' => [
        'name' => 'published',
        'label' => $this->t('Published'),
        'color' => '#4d6aff',
      ],
      'current' => [
        'name' => 'current',
        'label' => $this->t('Current'),
        'color' => '#ff6d1f',
      ],
      'live' => [
        'name' => 'live',
        'label' => $this->t('Live'),
        'color' => '#28a745',
      ],
      'latest' => [
        'name' => 'latest',
        'label' => $this->t('Latest revision'),
        'color' => '#681b98',
      ],
      'default' => [
        'name' => 'default',
        'label' => $this->t('Default'),
        'color' => '#0D142B',
      ],
    ];
    $commits = [];
    $node_storage = $this->entityTypeManager->getStorage('node');
    foreach ($revisions as $vid => $revision_data) {
      $commit = [
        'id' => (string) $vid,
        'author' => $revision_data['author'],
        'timestamp' => $revision_data['timestamp'],
        'urls' => $revision_data['urls'],
        'message' => $revision_data['message'],
        'branch' => $langcode,
        'parents' => [],
        'tags' => [],
      ];
      /** @var \Drupal\node\NodeInterface $revision */
      $revision = $node_storage->loadRevision($vid);
      $revision = $revision->getTranslation($langcode);
      // Subject can be a good candidate for moderation state.
      if ($revision_data['is_current_revision']) {
        $commit['tags'][] = $tags_list['current'];
      }
      if ($revision->isLatestRevision()) {
        $commit['tags'][] = $tags_list['latest'];
      }
      if ($revision->isPublished()) {
        $commit['tags'][] = $tags_list['published'];
        if ($revision_data['is_current_revision']) {
          $commit['tags'][] = $tags_list['live'];
        }
      }
      if ($revision->isDefaultRevision()) {
        $commit['tags'][] = $tags_list['default'];
      }
      $commits[(int) $vid] = $commit;
    }
    return $commits;
  }

}
